/*	primdraw.c by Lameguy64
	2014 Meido-Tek Productions.
	
	Demonstrates:
		- Using a primitive OT to draw triangles without libgs.
		- Using the GTE to rotate, translate, and project 3D primitives.
	
	Controls:
		Start							- Toggle interactive/non-interactive mode.
		Select							- Reset object's position and angles.
		L1/L2							- Move object closer/farther.
		L2/R2							- Rotate object (XY).
		Up/Down/Left/Right				- Rotate object (XZ/YZ).
		Triangle/Cross/Square/Circle	- Move object up/down/left/right.
		
*/

#include <sys/types.h>
#include <libgte.h>
#include <libgpu.h>
#include <libetc.h>
#include <stdio.h>

// Sample vector model
#include "trisamp.c"

#define HI_RES		1		// 0: 320x240, 1: 640x480

#define OT_LENGTH	1024	// Maximum number of OT entries
#define MAX_PRIMS	256		// Maximum number of POLY_FT3 primitives

// Center screen coordinates
#if HI_RES
#define CENTERX		320
#define CENTERY		240
#else
#define CENTERX		160
#define CENTERY		120
#endif

// Display and draw environments
DISPENV disp;
DRAWENV draw;

u_long	myOT[2][OT_LENGTH];		// Ordering table (contains addresses to primitives)
POLY_F3	myPrims[2][MAX_PRIMS];	// Primitive list
int		myPrimNum=0;			// Primitive counter


// Prototypes
int main();


int main() {
	
	int		ActivePage	=0;	
	#if HI_RES == 0
	RECT	ClearRect	={0,0,320,240};
	#else
	RECT	ClearRect	={0,0,640,480};
	#endif
	
	
	int		i;
	int		PadStatus;
	int		TPressed=0;
	int		AutoRotate=1;
	
	long	t, p, OTz, Flag;
	
	SVECTOR	Rotate={ 0 };					// Rotation coordinates
	VECTOR	Trans={ 0, 0, CENTERX, 0 };		// Translation coordinates
	
	// Scaling coordinates
	VECTOR	Scale={ ONE*(1+HI_RES), ONE*(1+HI_RES), ONE*(1+HI_RES), 0 };
	
	MATRIX	Matrix={0};						// Matrix data for the GTE
	
	
	// Reset the GPU before doing anything and the controller
	PadInit(0);
	ResetGraph(0);
	
	
	// Initialize and setup the GTE
	InitGeom();
	SetGeomOffset(CENTERX, CENTERY);
	SetGeomScreen(CENTERX);
	
	
	// Init font system
	FntLoad(960, 0);
	FntOpen(0, 0, 640, 480, 0, 512);
	
	
	// Set the display and draw environments
	#if HI_RES == 0
	SetDefDispEnv(&disp, 0, 0, 320, 240);
	SetDefDrawEnv(&draw, 0, 240, 320, 240);
	#else
	SetDefDispEnv(&disp, 0, 0, 640, 480);
	SetDefDrawEnv(&draw, 0, 0, 640, 480);	
	#endif
	
	
	// Set the new display/drawing environments
	VSync(0);
	PutDispEnv(&disp);
	PutDrawEnv(&draw);
	ClearImage(&ClearRect, 0, 0, 127);
	
	
	// Main loop
	while (1) {
	
		// Render the banner (FntPrint is always on top because it is not part of the OT)
		#if HI_RES
		FntPrint("\n\n");
		#endif
		FntPrint("\n\n SIMPLE 3D EXAMPLE BY LAMEGUY64\n");
		FntPrint(" 2014 MEIDO-TEK PRODUCTIONS\n");
		FntPrint(" WWW.PSXDEV.NET\n");
		
		
		// Read pad status
		PadStatus = PadRead(0);
		
		if (AutoRotate == 0) {
		
			if (PadStatus & PADL1) Trans.vz -= 4;
			if (PadStatus & PADR1) Trans.vz += 4;
			if (PadStatus & PADL2) Rotate.vz -= 8;
			if (PadStatus & PADR2) Rotate.vz += 8;

			if (PadStatus & PADLup)		Rotate.vx -= 8;
			if (PadStatus & PADLdown)	Rotate.vx += 8;
			if (PadStatus & PADLleft)	Rotate.vy -= 8;
			if (PadStatus & PADLright)	Rotate.vy += 8;
			
			if (PadStatus & PADRup)		Trans.vy -= 2;
			if (PadStatus & PADRdown)	Trans.vy += 2;
			if (PadStatus & PADRleft)	Trans.vx -= 2;
			if (PadStatus & PADRright)	Trans.vx += 2;
					
			if (PadStatus & PADselect) {
				Rotate.vx = Rotate.vy = Rotate.vz = 0;
				Scale.vx = Scale.vy = Scale.vz = ONE*(1+HI_RES);
				Trans.vx = Trans.vy = 0;
				Trans.vz = CENTERX;
			}
			
		}
		
		if (PadStatus & PADstart) {
			if (TPressed == 0) {
				AutoRotate = (AutoRotate + 1) & 1;
				Rotate.vx = Rotate.vy = Rotate.vz = 0;
				Scale.vx = Scale.vy = Scale.vz = ONE*(1+HI_RES);
				Trans.vx = Trans.vy = 0;
				Trans.vz = CENTERX;
			}
			TPressed = 1;
		} else {
			TPressed = 0;
		}

		if (AutoRotate) {
			Rotate.vy += 8;	// Pan
			Rotate.vx += 8;	// Tilt
			//Rotate.vz += 8;	// Roll
		}
		
		
		// Clear the current OT
		ClearOTagR(&myOT[ActivePage][0], OT_LENGTH);
		myPrimNum = 0;
		
		
		// Convert and set the matrixes
		RotMatrix(&Rotate, &Matrix);
		TransMatrix(&Matrix, &Trans);
		ScaleMatrix(&Matrix, &Scale);
		
		SetRotMatrix(&Matrix);
		SetTransMatrix(&Matrix);
		
		
		// Render the sample vector model
		t=0;
		for (i=0; i<vec.len; i++) {
			
			// Initialize the primitive and set its color values
			SetPolyF3(&myPrims[ActivePage][myPrimNum]);
			setRGB0(&myPrims[ActivePage][myPrimNum], vec_color[i].r, vec_color[i].g, vec_color[i].b);
			
			// Rotate, translate, and project the vectors and output the results into a primitive
			OTz = RotTransPers(&vec_mesh[t], (long*)&myPrims[ActivePage][myPrimNum].x0, &p, &Flag);
			OTz += RotTransPers(&vec_mesh[t+1], (long*)&myPrims[ActivePage][myPrimNum].x1, &p, &Flag);
			OTz += RotTransPers(&vec_mesh[t+2], (long*)&myPrims[ActivePage][myPrimNum].x2, &p, &Flag);
			
			// Sort the primitive into the OT
			OTz /= 3;
			if ((OTz > 0) && (OTz < OT_LENGTH))
				AddPrim(&myOT[ActivePage][OTz-2], &myPrims[ActivePage][myPrimNum]);
			
			myPrimNum++;
			t+=3;
			
		}
		
		
		// Prepare to switch buffers
		#if HI_RES == 0
		SetDefDispEnv(&disp, 0, 240*ActivePage, 320, 240);
		SetDefDrawEnv(&draw, 0, 240*(1-ActivePage), 320, 240);
		#endif
		

		// Wait for all drawing to finish
		DrawSync(0);
		
		// Clear the current buffer before rendering the next frame
		#if HI_RES == 0
		ClearRect.y = 240*ActivePage;
		#endif
		ClearImage(&ClearRect, 0, 0, 127);
		
		// Begin rendering the next frame
		DrawOTag(&myOT[ActivePage][OT_LENGTH-1]);
		FntFlush(0);
		
		
		// Wait for VSync and then switch buffers
		VSync(0);
		#if HI_RES == 0
		PutDispEnv(&disp);
		PutDrawEnv(&draw);
		#endif
		SetDispMask(1);
		
		
		// Toggle buffer index
		#if HI_RES == 0
		ActivePage = (ActivePage + 1) & 1;
		#endif
		
	}
	
}